"""Tests for the Mork API '/users/' endpoints."""

from uuid import uuid4

import pytest
from faker import Faker
from httpx import AsyncClient
from sqlalchemy import func, select

from mork.factories.users import (
    UserFactory,
    UserServiceStatusFactory,
)
from mork.models.users import (
    DeletionStatus,
    ServiceName,
    User,
    UserServiceStatus,
)


@pytest.mark.anyio
async def test_users_auth(http_client: AsyncClient):
    """Test required authentication for deletions endpoints."""
    # FastAPI returns a 403 error (instead of a 401 error) if no API token is given
    # see https://github.com/tiangolo/fastapi/discussions/9130
    assert (await http_client.get("/v1/users")).status_code == 403
    assert (await http_client.get("/v1/users/foo")).status_code == 403
    assert (await http_client.get("/v1/users/foo/status/bar")).status_code == 403
    assert (await http_client.patch("/v1/users/foo/status/bar")).status_code == 403


@pytest.mark.anyio
async def test_users_read_default(
    db_session, http_client: AsyncClient, auth_headers: dict
):
    """Test the behavior of retrieving the list of users to be deleted."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create 200 users that need to be deleted on all services
    number_users = 200
    UserFactory.create_batch(number_users)

    response = await http_client.get("/v1/users", headers=auth_headers)
    response_data = response.json()

    assert response.status_code == 200

    # Assert that the number of users matches the default limit (100)
    assert len(response_data) == 100

    # Assert that all users are unique
    assert len({user["id"] for user in response_data}) == len(response_data)

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize(
    "offset, limit, number_users",
    [
        (10, 100, 50),
        (0, 75, 100),
        (100, 75, 100),
        (100, 100, 200),
        (0, 0, 100),
        (50, 0, 10),
    ],
)
async def test_users_read_pagination(  # noqa: PLR0913
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
    offset: int,
    limit: int,
    number_users: int,
):
    """Test the pagination behavior of retrieving users."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some user requests in the database
    UserFactory.create_batch(number_users)

    # Assert the expected number of users have been created
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users

    # Get users with pagination
    response = await http_client.get(
        "/v1/users",
        headers=auth_headers,
        params={
            "deletion_status": "to_delete",
            "service": "ashley",
            "offset": offset,
            "limit": limit,
        },
    )
    response_data = response.json()
    assert response.status_code == 200

    # Assert that the number of users matches the pagination params
    expected_count = min(limit, max(0, number_users - offset))
    assert len(response_data) == expected_count

    # Assert that all user IDs are unique
    assert len({user["id"] for user in response_data}) == len(response_data)

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize(
    "invalid_params",
    [
        {"deletion_status": "wrong_status", "service": "ashley"},
        {"deletion_status": "to_delete", "service": "wrong_service"},
        {"deletion_status": "to_delete", "service": "ashley", "limit": 1001},
        {"deletion_status": "to_delete", "service": "ashley", "offset": -1},
    ],
)
async def test_users_read_invalid_params(
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
    invalid_params: dict,
):
    """Test scenarios with invalid query params when retrieving users."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users
    number_users = 10
    UserFactory.create_batch(number_users)

    # Read users with invalid query parameters
    response = await http_client.get(
        "/v1/users", headers=auth_headers, params=invalid_params
    )

    # Assert the request fails
    assert response.status_code == 422

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize(
    "params, expected_count",
    [
        ({"deletion_status": "to_delete"}, 5),
        ({"deletion_status": "deleted"}, 6),
        ({"deletion_status": "to_delete", "service": "ashley"}, 3),
        ({"deletion_status": "to_delete", "service": "joanie"}, 2),
        ({"deletion_status": "to_delete", "service": "brevo"}, 0),
    ],
)
async def test_users_read_filter(
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
    params: dict,
    expected_count: int,
):
    """Test scenarios with valid query parameters when retrieving users."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Attempt to retrieve users when the database is empty
    # with valid query parameters
    response = await http_client.get("/v1/users", headers=auth_headers, params=params)
    assert response.status_code == 200
    assert response.json() == []

    # Create 3 users to be deleted for Ashley
    UserFactory.create_batch(
        3,
        service_statuses={
            ServiceName.ASHLEY: DeletionStatus.TO_DELETE,
            ServiceName.BREVO: DeletionStatus.DELETED,
            ServiceName.EDX: DeletionStatus.DELETED,
            ServiceName.JOANIE: DeletionStatus.DELETED,
        },
    )
    # Create 2 users to be deleted for Joanie
    UserFactory.create_batch(
        2,
        service_statuses={
            ServiceName.ASHLEY: DeletionStatus.DELETED,
            ServiceName.BREVO: DeletionStatus.DELETED,
            ServiceName.EDX: DeletionStatus.DELETED,
            ServiceName.JOANIE: DeletionStatus.TO_DELETE,
        },
    )
    # Create 1 user deleted from all services
    UserFactory.create_batch(
        1,
        service_statuses={
            ServiceName.ASHLEY: DeletionStatus.DELETED,
            ServiceName.BREVO: DeletionStatus.DELETED,
            ServiceName.EDX: DeletionStatus.DELETED,
            ServiceName.JOANIE: DeletionStatus.DELETED,
        },
    )

    # Read users with valid query parameters
    response = await http_client.get(
        "/v1/users",
        headers=auth_headers,
        params=params,
    )

    response_data = response.json()
    assert response.status_code == 200

    # Verify we have the expected count
    assert len(response_data) == expected_count

    # Verify all IDs are unique
    assert len({user["id"] for user in response_data}) == expected_count

    # Verify the total count in database hasn't changed
    total_requests = (
        db_session.execute(
            select(User)
            .join(UserServiceStatus)
            .where(UserServiceStatus.status == DeletionStatus.TO_DELETE)
        )
        .scalars()
        .all()
    )
    assert len(total_requests) == 5


@pytest.mark.anyio
async def test_user_read(db_session, http_client: AsyncClient, auth_headers: dict):
    """Test the behavior of retrieving one user."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    creation_date = Faker().date_time()

    # Create one user that needs to be deleted on all services
    user = UserFactory.create(created_at=creation_date, updated_at=creation_date)

    # Get id of newly created user
    user_id = db_session.scalar(select(User.id))

    response = await http_client.get(f"/v1/users/{str(user_id)}", headers=auth_headers)

    assert response.status_code == 200

    # Verify the retrieved data matches the expected format
    assert response.json() == {
        "id": str(user_id),
        "username": user.username,
        "edx_user_id": user.edx_user_id,
        "email": user.email,
        "reason": user.reason.value,
        "service_statuses": [
            {
                "service_name": "ashley",
                "status": "to_delete",
            },
            {
                "service_name": "edx",
                "status": "to_delete",
            },
            {
                "service_name": "brevo",
                "status": "to_delete",
            },
            {
                "service_name": "joanie",
                "status": "to_delete",
            },
        ],
        "created_at": creation_date.isoformat(),
        "updated_at": creation_date.isoformat(),
    }


@pytest.mark.anyio
async def test_user_read_with_invalid_email(
    db_session, http_client: AsyncClient, auth_headers: dict
):
    """Test the behavior of retrieving one user with an invalid email address."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    creation_date = Faker().date_time()

    # Create one user with an invalid email, that needs to be deleted
    user = UserFactory.create(
        email="johndoe@example.com.",
        created_at=creation_date,
        updated_at=creation_date,
    )

    # Get id of newly created user
    user_id = db_session.scalar(select(User.id))

    response = await http_client.get(f"/v1/users/{str(user_id)}", headers=auth_headers)

    assert response.status_code == 200

    # Verify the retrieved data matches the expected format
    assert response.json() == {
        "id": str(user_id),
        "username": user.username,
        "edx_user_id": user.edx_user_id,
        "email": user.email,
        "reason": user.reason.value,
        "service_statuses": [
            {
                "service_name": "ashley",
                "status": "to_delete",
            },
            {
                "service_name": "edx",
                "status": "to_delete",
            },
            {
                "service_name": "brevo",
                "status": "to_delete",
            },
            {
                "service_name": "joanie",
                "status": "to_delete",
            },
        ],
        "created_at": creation_date.isoformat(),
        "updated_at": creation_date.isoformat(),
    }


@pytest.mark.anyio
@pytest.mark.parametrize("invalid_id", ["foo", 123, "a1-a2-aa", uuid4().hex + "a"])
async def test_user_read_invalid_id(
    db_session, http_client: AsyncClient, auth_headers: dict, invalid_id
):
    """Test the behavior of retrieving one user with an invalid id."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users in the database
    number_users = 10
    UserFactory.create_batch(number_users)

    # Attempt to read a user with an invalid ID
    response = await http_client.get(f"/v1/users/{invalid_id}", headers=auth_headers)

    assert response.status_code == 422

    # Attempt to read an user with a nonexistent ID
    nonexistent_id = uuid4().hex
    response = await http_client.get(
        f"/v1/users/{nonexistent_id}", headers=auth_headers
    )

    assert response.status_code == 404
    assert response.json() == {"detail": "User not found"}

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
async def test_user_read_invalid_params(
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
):
    """Test the behavior of retrieving one user with invalid query params."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user
    UserFactory.create()

    # Get id of newly created user
    user_id = db_session.scalar(select(User.id))

    # Read user with invalid query parameters
    response = await http_client.get(
        f"/v1/users/{user_id}",
        headers=auth_headers,
        params={"service": "wrong_service"},
    )

    # Assert the request fails
    assert response.status_code == 422

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == 1


@pytest.mark.anyio
async def test_user_read_status(
    db_session, http_client: AsyncClient, auth_headers: dict
):
    """Test the behavior of retrieving the status of a user."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user that need to be deleted on all services
    UserFactory.create()

    # Get id of newly created user
    user_id = db_session.scalar(select(User.id))

    response = await http_client.get(
        f"/v1/users/{str(user_id)}/status/ashley", headers=auth_headers
    )

    assert response.status_code == 200

    # Verify the retrieved data matches the expected format
    assert response.json() == {
        "id": str(user_id),
        "service_name": "ashley",
        "status": "to_delete",
    }


@pytest.mark.anyio
@pytest.mark.parametrize("invalid_id", ["foo", 123, "a1-a2-aa", uuid4().hex + "a"])
async def test_user_read_status_invalid_id(
    db_session, http_client: AsyncClient, auth_headers: dict, invalid_id
):
    """Test the behavior of retrieving a user status with an invalid user id."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users in the database
    number_users = 10
    UserFactory.create_batch(number_users)

    # Attempt to read a user with an invalid ID
    response = await http_client.get(
        f"/v1/users/{invalid_id}/status/ashley", headers=auth_headers
    )

    assert response.status_code == 422

    # Attempt to read an user with a nonexistent ID
    nonexistent_id = uuid4().hex
    response = await http_client.get(
        f"/v1/users/{nonexistent_id}/status/ashley", headers=auth_headers
    )

    assert response.status_code == 404
    assert response.json() == {"detail": "User status not found"}

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize("invalid_service", ["foo", 123])
async def test_user_read_status_invalid_service(
    db_session, http_client: AsyncClient, auth_headers: dict, invalid_service
):
    """Test the behavior of retrieving a user status with an invalid service name."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users in the database
    number_users = 10
    UserFactory.create_batch(number_users)

    # Get id of one of the newly created user
    user_id = db_session.scalar(select(User.id))

    # Attempt to read a user status with an invalid service
    response = await http_client.get(
        f"/v1/users/{user_id}/status/{invalid_service}", headers=auth_headers
    )

    assert response.status_code == 422

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize(
    "service_name, deletion_status",
    [
        ("ashley", "deleting"),
        ("brevo", "deleting"),
        ("edx", "deleted"),
        ("joanie", "deleted"),
    ],
)
async def test_users_update_status_default(
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
    service_name: str,
    deletion_status: str,
):
    """Test the behavior of updating the deletion status for a user on a service."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    user = UserFactory.create()

    # Assert user is to be deleted on the service
    status = (
        db_session.execute(
            select(UserServiceStatus.status).where(
                UserServiceStatus.service_name == service_name
            )
        )
        .scalars()
        .one()
    )
    assert status == DeletionStatus.TO_DELETE

    # Update the status of the user for a service
    response = await http_client.patch(
        f"/v1/users/{user.id}/status/{service_name}",
        headers=auth_headers,
        json={"deletion_status": deletion_status},
    )

    response_data = response.json()

    assert response.status_code == 200

    # Assert response is as expected
    assert response_data["id"] == str(user.id)
    assert response_data["service_name"] == service_name
    assert response_data["status"] == deletion_status

    # Assert that user status has correctly been updated
    updated_status = (
        db_session.execute(
            select(UserServiceStatus.status).where(
                UserServiceStatus.service_name == service_name
            )
        )
        .scalars()
        .one()
    )
    assert updated_status == deletion_status

    # Assert only this status has been updated
    number_updated = db_session.execute(
        select(func.count()).where(UserServiceStatus.status == deletion_status)
    ).scalar()
    assert number_updated == 1


@pytest.mark.anyio
@pytest.mark.parametrize("invalid_id", ["foo", 123, "a1-a2-aa", uuid4().hex + "a"])
async def test_user_update_invalid_id(
    db_session, http_client: AsyncClient, auth_headers: dict, invalid_id
):
    """Test the behavior of updating a user status for an invalid user id."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users in the database
    number_users = 10
    UserFactory.create_batch(number_users)

    # Attempt to update a user with an invalid ID
    response = await http_client.patch(
        f"/v1/users/{invalid_id}/status/ashley",
        headers=auth_headers,
        json={"deletion_status": "deleted"},
    )

    assert response.status_code == 422

    # Attempt to update a user with a nonexistent ID
    nonexistent_id = uuid4().hex
    response = await http_client.patch(
        f"/v1/users/{nonexistent_id}/status/ashley",
        headers=auth_headers,
        json={"deletion_status": "deleted"},
    )

    assert response.status_code == 404
    assert response.json() == {"detail": "User status not found"}

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize("invalid_service", ["foo", 123])
async def test_user_update_status_invalid_service(
    db_session, http_client: AsyncClient, auth_headers: dict, invalid_service
):
    """Test the behavior of updating a user status for an invalid service name."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create some users in the database
    number_users = 10
    UserFactory.create_batch(number_users)

    # Get id of one of the newly created user
    user_id = db_session.scalar(select(User.id))

    # Attempt to read a user status with an invalid service
    response = await http_client.patch(
        f"/v1/users/{user_id}/status/{invalid_service}", headers=auth_headers
    )

    assert response.status_code == 422

    # Assert the database still contains the same number of users
    users = db_session.execute(select(User)).all()
    assert len(users) == number_users


@pytest.mark.anyio
@pytest.mark.parametrize(
    "deletion_status",
    [
        "wrong_status",
        123,
    ],
)
async def test_users_update_status_invalid_status(
    db_session,
    http_client: AsyncClient,
    auth_headers: dict,
    deletion_status: str,
):
    """Test the behavior of updating the user status with an invalid status"""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    user = UserFactory.create()

    # Try to update status with invalid parameters
    response = await http_client.patch(
        f"/v1/users/{user.id}/status/ashley",
        headers=auth_headers,
        json={"deletion_status": deletion_status},
    )

    # Assert the request fails
    assert response.status_code == 422



"""Mork users schemas."""

from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, ConfigDict

from mork.models.users import DeletionReason, DeletionStatus, ServiceName


class UserServiceStatusRead(BaseModel):
    """Model for reading service statuses of a user."""

    model_config = ConfigDict(from_attributes=True)

    service_name: ServiceName
    status: DeletionStatus


class UserRead(BaseModel):
    """Model for reading detailed information about a user."""

    model_config = ConfigDict(from_attributes=True)

    id: UUID
    username: str
    edx_user_id: int
    email: str
    reason: DeletionReason
    service_statuses: list[UserServiceStatusRead]
    created_at: datetime
    updated_at: datetime


class UserStatusRead(BaseModel):
    """Model for reading a user status."""

    model_config = ConfigDict(from_attributes=True)

    id: UUID
    service_name: ServiceName
    status: DeletionStatus


class UserStatusUpdate(BaseModel):
    """Model for response after updating a user status."""

    model_config = ConfigDict(from_attributes=True)

    id: UUID
    service_name: ServiceName
    status: DeletionStatus



"""Celery utils functions."""

from logging import getLogger
from uuid import UUID

import httpx

from mork.conf import settings
from mork.models.users import DeletionStatus, ServiceName
from mork.schemas.users import UserRead

logger = getLogger(__name__)


def get_user_from_mork(user_id: UUID) -> UserRead | None:
    """Retrieve user from Mork by ID."""
    logger.debug("Get user from Mork")
    logger.debug(f"API URL: {settings.SERVER_URL}")

    try:
        response = httpx.get(
            f"{settings.SERVER_URL}/v1/users/{user_id}",
            headers={"X-API-Key": f"{settings.API_KEYS[0]}"},
        )
        response.raise_for_status()
    except httpx.HTTPError as exc:
        logger.error(f"Failed to retrieve user from Mork API: {exc}")
        return None

    return UserRead.model_validate(response.json())


def get_service_status(user: UserRead, service: ServiceName) -> DeletionStatus | None:
    """Find the service status entry for a user."""
    service_status = next(
        (status for status in user.service_statuses if status.service_name == service),
        None,
    )
    return service_status.status if service_status else None


def update_status_in_mork(
    user_id: UUID, service: ServiceName, status: DeletionStatus
) -> bool:
    """Update the user deletion status in Mork."""
    logger.debug(f"Updating deletion status for user {user_id} in Mork to {status}")
    logger.debug(f"API URL: {settings.SERVER_URL}")

    try:
        response = httpx.patch(
            f"{settings.SERVER_URL}/v1/users/{user_id}/status/{service.value}",
            headers={"X-API-Key": f"{settings.API_KEYS[0]}"},
            json={"deletion_status": status.value},
        )
        response.raise_for_status()
    except httpx.HTTPError as exc:
        logger.error(f"Failed to update user status with Mork API: {exc}")
        return False

    return True



"""Tests for Mork Celery emailing tasks."""

from unittest.mock import Mock, call

import pytest
from faker import Faker

from mork.celery.tasks.emailing import (
    check_email_already_sent,
    mark_email_status,
    warn_inactive_users,
    warn_user,
)
from mork.conf import settings
from mork.edx.mysql.factories.auth import EdxAuthUserFactory
from mork.exceptions import EmailSendError
from mork.factories.tasks import EmailStatusFactory


def test_warn_inactive_users(edx_mysql_db, monkeypatch):
    """Test the `warn_inactive_users` function."""
    # 2 users that did not log in for more than the warning period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe1",
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe2",
        email="johndoe2@example.com",
    )
    # 2 users that logged in recently
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(start_date=-settings.WARNING_PERIOD),
        username="JaneDah1",
        email="janedah1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(start_date=-settings.WARNING_PERIOD),
        username="JaneDah2",
        email="janedah2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.emailing.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.group", mock_group)
    mock_warn_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.warn_user", mock_warn_user)

    warn_inactive_users(dry_run=False)

    mock_group.assert_called_once_with(
        [
            mock_warn_user.s(
                email="johndoe1@example.com", username="JohnDoe1", dry_run=False
            ),
            mock_warn_user.s(
                email="johndoe2@example.com", username="JohnDoe2", dry_run=False
            ),
        ]
    )


def test_warn_inactive_users_with_limit(edx_mysql_db, monkeypatch):
    """Test the `warn_inactive_users` function with limit."""
    # 2 users that did not log in for more than the warning period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe1",
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe2",
        email="johndoe2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.emailing.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.group", mock_group)
    mock_warn_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.warn_user", mock_warn_user)

    warn_inactive_users(limit=1, dry_run=False)

    mock_group.assert_called_once_with(
        [
            mock_warn_user.s(
                email="johndoe1@example.com", username="JohnDoe1", dry_run=False
            )
        ]
    )


def test_warn_inactive_users_with_batch_size(edx_mysql_db, monkeypatch):
    """Test the `warn_inactive_users` function with batch size."""
    # 2 users that did not log in for more than the warning period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe1",
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe2",
        email="johndoe2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.emailing.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.group", mock_group)
    mock_warn_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.warn_user", mock_warn_user)

    # Set batch size to 1
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.settings.EDX_MYSQL_QUERY_BATCH_SIZE", 1
    )

    warn_inactive_users(dry_run=False)

    mock_group.assert_has_calls(
        [
            call(
                [
                    mock_warn_user.s(
                        email="johndoe1@example.com", username="JohnDoe1", dry_run=False
                    ),
                ]
            ),
            call().delay(),
            call(
                [
                    mock_warn_user.s(
                        email="johndoe2@example.com", username="JohnDoe2", dry_run=False
                    ),
                ]
            ),
            call().delay(),
        ]
    )


def test_warn_inactive_users_with_dry_run(edx_mysql_db, monkeypatch):
    """Test the `warn_inactive_users` function with dry run activated (by default)."""
    # 2 users that did not log in for more than the warning period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe1",
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.WARNING_PERIOD),
        username="JohnDoe2",
        email="johndoe2@example.com",
    )
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.group", mock_group)
    mock_warn_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.warn_user", mock_warn_user)

    warn_inactive_users()

    mock_group.assert_called_once_with(
        [
            mock_warn_user.s(
                email="johndoe1@example.com", username="JohnDoe1", dry_run=True
            ),
            mock_warn_user.s(
                email="johndoe2@example.com", username="JohnDoe2", dry_run=True
            ),
        ]
    )


def test_warn_user(monkeypatch):
    """Test the `warn_user` function."""
    mock_check_email = Mock(return_value=False)
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.check_email_already_sent", mock_check_email
    )
    mock_send_email = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.send_email", mock_send_email)
    mock_mark_email_status = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.mark_email_status", mock_mark_email_status
    )

    email = "johndoe@example.com"
    username = "JohnDoe"
    warn_user(email, username, dry_run=False)

    mock_check_email.assert_called_once_with(email)
    mock_send_email.assert_called_once_with(email, username)
    mock_mark_email_status.assert_called_once_with(email)


def test_warn_user_with_dry_run(monkeypatch):
    """Test the `warn_user` function with dry run activated (by default)."""
    mock_check_email = Mock(return_value=False)
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.check_email_already_sent", mock_check_email
    )
    mock_send_email = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.send_email", mock_send_email)
    mock_mark_email_status = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.mark_email_status", mock_mark_email_status
    )

    email = "johndoe@example.com"
    username = "JohnDoe"
    warn_user(email, username)

    mock_check_email.assert_called_once_with(email)
    mock_send_email.assert_not_called()
    mock_mark_email_status.assert_not_called()


def test_warn_user_already_sent(monkeypatch):
    """Test the `warn_user` function when email has already been sent."""
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.check_email_already_sent", lambda x: True
    )

    mock_send_email = Mock()
    monkeypatch.setattr("mork.celery.tasks.emailing.send_email", mock_send_email)

    warn_user("johndoe@example.com", "JohnDoe")

    mock_send_email.assert_not_called()


def test_warn_user_sending_failure(monkeypatch):
    """Test the `warn_user` function with email sending failure."""
    monkeypatch.setattr(
        "mork.celery.tasks.emailing.check_email_already_sent", lambda x: False
    )

    def mock_send(*args):
        raise EmailSendError("An error occurred")

    monkeypatch.setattr("mork.celery.tasks.emailing.send_email", mock_send)

    with pytest.raises(EmailSendError, match="An error occurred"):
        warn_user("johndoe@example.com", "JohnDoe", dry_run=False)


def test_check_email_already_sent(monkeypatch, db_session):
    """Test the `check_email_already_sent` function."""
    email_address = "test_email@example.com"

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.emailing.MorkDB", MockMorkDB)
    EmailStatusFactory.create_batch(3)

    assert not check_email_already_sent(email_address)

    EmailStatusFactory.create(email=email_address)
    assert check_email_already_sent(email_address)


def test_mark_email_status(monkeypatch, db_session):
    """Test the `mark_email_status` function."""

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.emailing.MorkDB", MockMorkDB)

    # Write new email status entry
    new_email = "test_email@example.com"
    mark_email_status(new_email)
    assert check_email_already_sent(new_email)



"""Factory classes for payment models."""

import factory

from mork.edx.mysql.models.payment import PaymentUseracceptance

from .base import BaseSQLAlchemyModelFactory


class EdxPaymentUseracceptanceFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `payment_useracceptance` table."""

    class Meta:
        """Factory configuration."""

        model = PaymentUseracceptance

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    terms_id = factory.Sequence(lambda n: n + 1)
    datetime = factory.Faker("date_time")



"""Mork Celery emailing tasks."""

from datetime import datetime
from logging import getLogger

from celery import group
from sqlalchemy import select

from mork.celery.celery_app import app
from mork.conf import settings
from mork.db import MorkDB
from mork.edx.mysql import crud
from mork.edx.mysql.database import OpenEdxMySQLDB
from mork.exceptions import EmailSendError
from mork.mail import send_email
from mork.models.tasks import EmailStatus

logger = getLogger(__name__)


@app.task
def warn_inactive_users(limit: int = 0, dry_run: bool = True):
    """Celery task to warn inactive users by email."""
    db = OpenEdxMySQLDB()

    threshold_date = datetime.now() - settings.WARNING_PERIOD
    total = crud.get_inactive_users_count(db.session, threshold_date)

    if limit:
        total = min(total, limit)

    for batch_offset in range(0, total, settings.EDX_MYSQL_QUERY_BATCH_SIZE):
        batch_limit = min(settings.EDX_MYSQL_QUERY_BATCH_SIZE, total - batch_offset)
        inactive_users = crud.get_inactive_users(
            db.session,
            threshold_date,
            offset=batch_offset,
            limit=batch_limit,
        )
        send_email_group = group(
            [
                warn_user.s(email=user.email, username=user.username, dry_run=dry_run)
                for user in inactive_users
            ]
        )
        send_email_group.delay()


@app.task(
    bind=True,
    rate_limit=settings.EMAIL_RATE_LIMIT,
    retry_kwargs={"max_retries": settings.EMAIL_MAX_RETRIES},
)
def warn_user(self, email: str, username: str, dry_run: bool = True):
    """Celery task that warns the specified user by sending an email."""
    # Check that user has not already received a warning email
    if check_email_already_sent(email):
        logger.warning("An email has already been sent to this user")
        return

    if dry_run:
        logger.info("Dry run: An email would have been sent")
        return

    logger.debug(f"Sending an email to user with {email=}")
    try:
        send_email(email, username)
    except EmailSendError as exc:
        logger.exception(exc)
        raise self.retry(exc=exc) from exc

    # Write flag that email was correctly sent to this user
    mark_email_status(email)


def check_email_already_sent(email: str):
    """Check if an email has already been sent to the user."""
    db = MorkDB()
    query = select(EmailStatus.email).where(EmailStatus.email == email)
    result = db.session.execute(query).scalars().first()
    db.session.close()
    return result


def mark_email_status(email: str):
    """Mark the email status in database."""
    db = MorkDB()
    db.session.add(EmailStatus(email=email, sent_date=datetime.now()))
    db.session.commit()
    db.session.close()



"""Factory classes for bulk models."""

import factory

from mork.edx.mysql.models.bulk import BulkEmailCourseemail, BulkEmailOptout

from .base import BaseSQLAlchemyModelFactory, faker


class EdxBulkEmailCourseemailFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `bulk_email_courseemail` table."""

    class Meta:
        """Factory configuration."""

        model = BulkEmailCourseemail

    id = factory.Sequence(lambda n: n + 1)
    sender_id = factory.Sequence(lambda n: n + 1)
    slug = factory.Faker("pystr")
    subject = factory.Faker("pystr")
    html_message = factory.Faker("pystr")
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    to_option = factory.Faker("word")
    text_message = factory.Faker("text")
    template_name = factory.Faker("text")
    from_addr = factory.Faker("text")


class EdxBulkEmailOptoutFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `bulk_email_optout` table."""

    class Meta:
        """Factory configuration."""

        model = BulkEmailOptout

    id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    user_id = factory.Sequence(lambda n: n + 1)



"""Mork edx models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class ContentstoreVideouploadconfig(Base):
    """Model for the `contentstore_videouploadconfig` table."""

    __tablename__ = "contentstore_videouploadconfig"
    __table_args__ = (
        ForeignKeyConstraint(
            ["changed_by_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    change_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    changed_by_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    enabled: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    profile_whitelist: Mapped[str] = mapped_column(TEXT, nullable=False)

    changed_by: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="contentstore_videouploadconfig"
    )



"""Mork edx models."""

from sqlalchemy.orm import DeclarativeBase


class Base(DeclarativeBase):
    """Base class for all models in the database."""

    filtered_attrs = []

    def safe_dict(self):
        """Return a dictionary representation of the model."""
        return {
            c.name: getattr(self, c.name)
            for c in self.__table__.columns
            if c.name not in self.filtered_attrs
        }



"""Main module for Mork API."""

from functools import lru_cache
from urllib.parse import urlparse

import sentry_sdk
from fastapi import FastAPI
from fastapi.concurrency import asynccontextmanager
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.starlette import StarletteIntegration

from mork import __version__
from mork.api.health import router as health_router
from mork.api.v1 import app as v1
from mork.conf import settings
from mork.db import get_engine


@lru_cache(maxsize=None)
def get_health_check_routes() -> list:
    """Return the health check routes."""
    return [route.path for route in health_router.routes]


def filter_transactions(event: dict, hint) -> dict | None:  # noqa: ARG001
    """Filter transactions for Sentry."""
    url = urlparse(event["request"]["url"])

    if settings.SENTRY_IGNORE_HEALTH_CHECKS and url.path in get_health_check_routes():
        return None

    return event


@asynccontextmanager
async def lifespan(app: FastAPI):  # noqa: ARG001
    """Application life span."""
    engine = get_engine()

    # Sentry
    if settings.SENTRY_DSN is not None:
        sentry_sdk.init(
            dsn=settings.SENTRY_DSN,
            enable_tracing=True,
            traces_sample_rate=settings.SENTRY_API_TRACES_SAMPLE_RATE,
            release=__version__,
            environment=settings.SENTRY_EXECUTION_ENVIRONMENT,
            max_breadcrumbs=50,
            before_send_transaction=filter_transactions,
            integrations=[
                StarletteIntegration(),
                FastApiIntegration(),
            ],
        )
        sentry_sdk.set_tag("application", "api")

    yield
    engine.dispose()


app = FastAPI(title="Mork", lifespan=lifespan)

# Health checks
app.include_router(health_router)

# Mount v1 API
app.mount("/v1", v1)



"""Mork edx student models."""

import datetime
from typing import List, Optional

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class StudentAnonymoususerid(Base):
    """Model for the `student_anonymoususerid` table."""

    __tablename__ = "student_anonymoususerid"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    anonymous_user_id: Mapped[str] = mapped_column(
        String(32), nullable=False, unique=True
    )
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_anonymoususerid"
    )


class StudentCourseaccessrole(Base):
    """Model for the `student_courseaccessrole` table."""

    __tablename__ = "student_courseaccessrole"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "student_courseaccessrole_user_id_3203176c4f474414_uniq",
            "user_id",
            "org",
            "course_id",
            "role",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    org: Mapped[str] = mapped_column(String(64), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    role: Mapped[str] = mapped_column(String(64), nullable=False, index=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_courseaccessrole"
    )


class StudentCourseenrollment(Base):
    """Model for the `student_courseenrollment` table."""

    __tablename__ = "student_courseenrollment"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
            name="user_id_refs_id_45948fcded37bc9d",
        ),
        Index("student_courseenrollment_3216ff68", "created"),
        Index("student_courseenrollment_fbfc09f1", "user_id"),
        Index("student_courseenrollment_ff48d8e5", "course_id"),
        Index(
            "student_courseenrollment_user_id_2d2a572f07dd8e37_uniq",
            "user_id",
            "course_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11))
    course_id: Mapped[str] = mapped_column(String(255))
    is_active: Mapped[int] = mapped_column(INTEGER(1))
    mode: Mapped[str] = mapped_column(String(100))
    created: Mapped[Optional[datetime.datetime]] = mapped_column(DateTime)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_courseenrollment"
    )

    student_courseenrollmentattribute: Mapped[
        List["StudentCourseenrollmentattribute"]
    ] = relationship(
        "StudentCourseenrollmentattribute",
        back_populates="enrollment",
        cascade="all, delete-orphan",
    )
    student_manualenrollmentaudit: Mapped[List["StudentManualenrollmentaudit"]] = (
        relationship(
            "StudentManualenrollmentaudit",
            back_populates="enrollment",
            cascade="all, delete-orphan",
        )
    )


class StudentCourseenrollmentallowed(Base):
    """Model for the `student_courseenrollmentallowed` table.

    At the database level, no foreign key is defined.
    """

    __tablename__ = "student_courseenrollmentallowed"
    __table_args__ = (
        Index(
            "student_courseenrollmentallowed_email_6f3eafd4a6c58591_uniq",
            "email",
            "course_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    email: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, index=True)
    auto_enroll: Mapped[int] = mapped_column(INTEGER(1), nullable=False)


class StudentCourseenrollmentattribute(Base):
    """Model for the `student_courseenrollmentattribute` table."""

    __tablename__ = "student_courseenrollmentattribute"
    __table_args__ = (
        ForeignKeyConstraint(
            ["enrollment_id"],
            ["student_courseenrollment.id"],
        ),
    )
    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    enrollment_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    namespace: Mapped[str] = mapped_column(String(255), nullable=False)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    value: Mapped[str] = mapped_column(String(255), nullable=False)

    enrollment: Mapped["StudentCourseenrollment"] = relationship(
        "StudentCourseenrollment", back_populates="student_courseenrollmentattribute"
    )


class StudentHistoricalcourseenrollment(Base):
    """Model for the `student_historicalcourseenrollment` table."""

    __tablename__ = "student_historicalcourseenrollment"
    __table_args__ = (
        ForeignKeyConstraint(
            ["history_user_id"],
            ["auth_user.id"],
        ),
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, index=True)
    is_active: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    mode: Mapped[str] = mapped_column(String(100), nullable=False)
    user_id: Mapped[int] = mapped_column(INTEGER(11), index=True, nullable=True)
    history_id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    history_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    history_user_id: Mapped[int] = mapped_column(INTEGER(11), index=True, nullable=True)
    history_type: Mapped[str] = mapped_column(String(1), nullable=False)

    history_user = relationship(
        "AuthUser",
        foreign_keys=[history_user_id],
        primaryjoin="StudentHistoricalcourseenrollment.history_user_id == AuthUser.id",
        back_populates="student_historicalcourseenrollment_history_user",
    )
    user = relationship(
        "AuthUser",
        foreign_keys=[user_id, history_user_id],
        primaryjoin="StudentHistoricalcourseenrollment.user_id == AuthUser.id",
        back_populates="student_historicalcourseenrollment_user",
    )


class StudentLanguageproficiency(Base):
    """Model for the `student_languageproficiency` table."""

    __tablename__ = "student_languageproficiency"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_profile_id"],
            ["auth_userprofile.id"],
        ),
        Index(
            "student_languageproficiency_code_68e76171684c62e5_uniq",
            "code",
            "user_profile_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_profile_id: Mapped[int] = mapped_column(
        INTEGER(11), nullable=False, index=True
    )
    code: Mapped[str] = mapped_column(String(16), nullable=False)

    user_profile: Mapped["AuthUserprofile"] = relationship(  # noqa: F821
        "AuthUserprofile", back_populates="student_languageproficiency"
    )


class StudentLoginfailure(Base):
    """Model for the `student_loginfailures` table."""

    __tablename__ = "student_loginfailures"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    failure_count: Mapped[int] = mapped_column(INTEGER(11), nullable=False)
    lockout_until: Mapped[datetime.datetime] = mapped_column(DateTime)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_loginfailures"
    )


class StudentManualenrollmentaudit(Base):
    """Model for the `student_manualenrollmentaudit` table."""

    __tablename__ = "student_manualenrollmentaudit"
    __table_args__ = (
        ForeignKeyConstraint(
            ["enrollment_id"],
            ["student_courseenrollment.id"],
        ),
        ForeignKeyConstraint(
            ["enrolled_by_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    enrollment_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    enrolled_by_id: Mapped[int] = mapped_column(INTEGER(11), nullable=True, index=True)
    enrolled_email: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    time_stamp: Mapped[datetime.datetime] = mapped_column(DateTime)
    state_transition: Mapped[str] = mapped_column(String(255), nullable=False)
    reason: Mapped[str] = mapped_column(TEXT)

    enrollment: Mapped["StudentCourseenrollment"] = relationship(
        "StudentCourseenrollment", back_populates="student_manualenrollmentaudit"
    )
    enrolled_by: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_manualenrollmentaudit"
    )


class StudentPendingemailchange(Base):
    """Model for the `student_pendingemailchange` table."""

    __tablename__ = "student_pendingemailchange"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, unique=True)
    new_email: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    activation_key: Mapped[str] = mapped_column(String(32), nullable=False, unique=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="student_pendingemailchange"
    )


class StudentUserstanding(Base):
    """Model for the `student_userstanding` table."""

    __tablename__ = "student_userstanding"
    __table_args__ = (
        ForeignKeyConstraint(
            ["changed_by_id"],
            ["auth_user.id"],
        ),
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, unique=True)
    account_status: Mapped[str] = mapped_column(String(31), nullable=False)
    changed_by_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    standing_last_changed_at: Mapped[datetime.datetime] = mapped_column(
        DateTime, nullable=False
    )

    changed_by = relationship(
        "AuthUser",
        foreign_keys=[changed_by_id],
        primaryjoin="StudentUserstanding.changed_by_id == AuthUser.id",
        back_populates="student_userstanding_changed_by",
    )
    user = relationship(
        "AuthUser",
        foreign_keys=[user_id, changed_by_id],
        primaryjoin="StudentUserstanding.user_id == AuthUser.id",
        back_populates="student_userstanding_user",
    )



"""Mork tasks schemas."""

from enum import Enum, unique
from typing import Literal

from pydantic import BaseModel, ConfigDict, EmailStr, PositiveInt

from mork.celery.tasks.deletion import delete_inactive_users, delete_user
from mork.celery.tasks.emailing import warn_inactive_users, warn_user


@unique
class TaskStatus(str, Enum):
    """Task statuses."""

    FAILURE = "FAILURE"
    PENDING = "PENDING"
    RECEIVED = "RECEIVED"
    RETRY = "RETRY"
    REVOKED = "REVOKED"
    STARTED = "STARTED"
    SUCCESS = "SUCCESS"


@unique
class TaskType(str, Enum):
    """Possible task types."""

    EMAIL_INACTIVE_USERS = "email_inactive_users"
    EMAIL_USER = "email_user"
    DELETE_INACTIVE_USERS = "delete_inactive_users"
    DELETE_USER = "delete_user"


class TaskCreateBase(BaseModel):
    """Base model for creating a task."""

    model_config = ConfigDict(extra="ignore")
    dry_run: bool = True


class DeleteInactiveUsers(TaskCreateBase):
    """Model for creating a task to delete all inactive users."""

    type: Literal[TaskType.DELETE_INACTIVE_USERS]
    limit: PositiveInt | None = None


class EmailInactiveUsers(TaskCreateBase):
    """Model for creating a task to email all inactive users."""

    type: Literal[TaskType.EMAIL_INACTIVE_USERS]
    limit: PositiveInt | None = None


class DeleteUser(TaskCreateBase):
    """Model for creating a task to delete one user."""

    type: Literal[TaskType.DELETE_USER]
    email: str


class EmailUser(TaskCreateBase):
    """Model for creating a task to email one user."""

    type: Literal[TaskType.EMAIL_USER]
    email: EmailStr
    username: str


class TaskResponse(BaseModel):
    """Model for a task response."""

    id: str
    status: TaskStatus


TASK_TYPE_TO_FUNC = {
    TaskType.EMAIL_INACTIVE_USERS: warn_inactive_users,
    TaskType.EMAIL_USER: warn_user,
    TaskType.DELETE_INACTIVE_USERS: delete_inactive_users,
    TaskType.DELETE_USER: delete_user,
}



"""Factory classes for user models."""

import factory

from mork.edx.mysql.models.user import UserApiUserpreference

from .base import BaseSQLAlchemyModelFactory


class EdxUserApiUserpreferenceFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `user_api_userpreference` table."""

    class Meta:
        """Factory configuration."""

        model = UserApiUserpreference

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    value = factory.Faker("pystr")

    @factory.lazy_attribute
    def key(self):
        """Pick a random key from the complete list of Open edX user preferences."""
        return ["account_privacy", "dark-lang", "pref-lang"][self.id % 3]



"""Factory classes for auth models."""

import factory

from mork.edx.mysql.models.auth import (
    AuthRegistration,
    AuthtokenToken,
    AuthUser,
    AuthUserGroups,
    AuthUserprofile,
)

from .base import BaseSQLAlchemyModelFactory, faker
from .bulk import EdxBulkEmailCourseemailFactory, EdxBulkEmailOptoutFactory
from .certificates import (
    EdxCertificatesCertificatehtmlviewconfigurationFactory,
    EdxCertificatesGeneratedCertificateFactory,
)
from .contentstore import EdxContentstoreVideouploadconfigFactory
from .course import (
    EdxCourseActionStateCoursererunstateFactory,
    EdxCourseCreatorsCoursecreatorFactory,
    EdxCourseGroupsCohortmembershipFactory,
    EdxCourseGroupsCourseusergroupUsersFactory,
)
from .courseware import (
    EdxCoursewareOfflinecomputedgradeFactory,
    EdxCoursewareStudentmoduleFactory,
    EdxCoursewareXmodulestudentinfofieldFactory,
    EdxCoursewareXmodulestudentprefsfieldFactory,
)
from .dark import EdxDarkLangDarklangconfigFactory
from .django import EdxDjangoCommentClientRoleUsersFactory
from .instructor import EdxInstructorTaskInstructortaskFactory
from .notify import EdxNotifySettingFactory
from .payment import EdxPaymentUseracceptanceFactory
from .proctoru import EdxProctoruProctoruexamFactory, EdxProctoruProctoruuserFactory
from .student import (
    EdxStudentAnonymoususeridFactory,
    EdxStudentCourseaccessroleFactory,
    EdxStudentCourseenrollmentFactory,
    EdxStudentHistoricalcourseenrollmentFactory,
    EdxStudentLanguageproficiencyFactory,
    EdxStudentLoginfailureFactory,
    EdxStudentManualenrollmentauditFactory,
    EdxStudentPendingemailchangeFactory,
    EdxStudentUserstandingFactory,
)
from .user import EdxUserApiUserpreferenceFactory
from .util import EdxUtilRatelimitconfigurationFactory
from .verify import (
    EdxVerifyStudentHistoricalverificationdeadlineFactory,
    EdxVerifyStudentSoftwaresecurephotoverificationFactory,
)
from .wiki import WikiArticleFactory, WikiArticlerevisionFactory


class EdxAuthRegistrationFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `auth_registration` table."""

    class Meta:
        """Factory configuration."""

        model = AuthRegistration

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    activation_key = factory.Faker("hexify", text="^" * 32)


class EdxAuthtokenTokenFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `authtoken_token` table."""

    class Meta:
        """Factory configuration."""

        model = AuthtokenToken

    key = factory.Faker("hexify", text="^" * 40)
    user_id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")


class EdxAuthUserGroupsFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `auth_user_groups` table."""

    class Meta:
        """Factory configuration."""

        model = AuthUserGroups

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    group_id = factory.Sequence(lambda n: n + 1)


class EdxAuthUserprofileFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `auth_userprofile` table."""

    class Meta:
        """Factory configuration."""

        model = AuthUserprofile

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    name = factory.Faker("name")
    language = factory.Faker("language_code")
    location = factory.Faker("address")
    meta = factory.Faker("json")
    courseware = factory.Faker("pystr")
    allow_certificate = factory.Faker("random_int", min=0, max=1)
    gender = factory.Faker("random_element", elements=["f", "m"])
    mailing_address = factory.Faker("email")
    year_of_birth = factory.Faker("year")
    level_of_education = factory.Faker("pystr", max_chars=6)
    goals = factory.Faker("text")
    country = factory.Faker("country_code")
    city = factory.Faker("city")
    bio = factory.Faker("text")
    profile_image_uploaded_at = factory.Faker("date_time")

    student_languageproficiency = factory.RelatedFactoryList(
        EdxStudentLanguageproficiencyFactory,
        "user_profile",
        size=3,
        user_profile_id=factory.SelfAttribute("..id"),
    )


class EdxAuthUserFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `auth_user` table."""

    class Meta:
        """Factory configuration."""

        model = AuthUser

    class Params:
        """Factory parameter to toggle generation of protected tables on or off."""

        with_protected_tables = factory.Trait(
            authtoken_token=factory.RelatedFactory(
                EdxAuthtokenTokenFactory, user_id=factory.SelfAttribute("..id")
            ),
            certificates_certificatehtmlviewconfiguration=factory.RelatedFactoryList(
                EdxCertificatesCertificatehtmlviewconfigurationFactory,
                "changed_by",
                size=3,
                changed_by_id=factory.SelfAttribute("..id"),
            ),
            contentstore_videouploadconfig=factory.RelatedFactoryList(
                EdxContentstoreVideouploadconfigFactory,
                "changed_by",
                size=3,
                changed_by_id=factory.SelfAttribute("..id"),
            ),
            course_action_state_coursererunstate_created_user=factory.RelatedFactoryList(
                EdxCourseActionStateCoursererunstateFactory,
                "created_user",
                size=3,
                created_user_id=factory.SelfAttribute("..id"),
            ),
            course_action_state_coursererunstate_updated_user=factory.RelatedFactoryList(
                EdxCourseActionStateCoursererunstateFactory,
                "updated_user",
                size=3,
                updated_user_id=factory.SelfAttribute("..id"),
            ),
            course_creators_coursecreator=factory.RelatedFactory(
                EdxCourseCreatorsCoursecreatorFactory,
                user_id=factory.SelfAttribute("..id"),
            ),
            dark_lang_darklangconfig=factory.RelatedFactoryList(
                EdxDarkLangDarklangconfigFactory,
                "changed_by",
                size=3,
                changed_by_id=factory.SelfAttribute("..id"),
            ),
            util_ratelimitconfiguration=factory.RelatedFactoryList(
                EdxUtilRatelimitconfigurationFactory,
                "changed_by",
                size=3,
                changed_by_id=factory.SelfAttribute("..id"),
            ),
            verify_student_historicalverificationdeadline=factory.RelatedFactoryList(
                EdxVerifyStudentHistoricalverificationdeadlineFactory,
                "history_user",
                size=3,
                history_user_id=factory.SelfAttribute("..id"),
            ),
            wiki_article=factory.RelatedFactoryList(
                WikiArticleFactory,
                "owner",
                size=3,
                owner_id=factory.SelfAttribute("..id"),
            ),
            wiki_articlerevision=factory.RelatedFactoryList(
                WikiArticlerevisionFactory,
                "user",
                size=3,
                user_id=factory.SelfAttribute("..id"),
            ),
        )

    id = factory.Sequence(lambda n: n + 1)
    username = factory.Sequence(lambda n: f"{faker.user_name()}{n}")
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    password = factory.Faker("password")
    email = factory.Faker("email")
    is_staff = False
    is_superuser = False
    is_active = True
    date_joined = factory.Faker("date_time")
    last_login = factory.Faker("date_time")

    auth_registration = factory.RelatedFactory(
        EdxAuthRegistrationFactory, user_id=factory.SelfAttribute("..id")
    )
    auth_userprofile = factory.RelatedFactory(
        EdxAuthUserprofileFactory, user_id=factory.SelfAttribute("..id")
    )
    auth_user_groups = factory.RelatedFactoryList(
        EdxAuthUserGroupsFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    bulk_email_courseemail = factory.RelatedFactoryList(
        EdxBulkEmailCourseemailFactory,
        "sender",
        size=3,
        sender_id=factory.SelfAttribute("..id"),
    )
    bulk_email_optout = factory.RelatedFactoryList(
        EdxBulkEmailOptoutFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    certificates_generatedcertificate = factory.RelatedFactoryList(
        EdxCertificatesGeneratedCertificateFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    course_groups_courseusergroup_users = factory.RelatedFactoryList(
        EdxCourseGroupsCourseusergroupUsersFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    course_groups_cohortmembership = factory.RelatedFactoryList(
        EdxCourseGroupsCohortmembershipFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    courseware_offlinecomputedgrade = factory.RelatedFactoryList(
        EdxCoursewareOfflinecomputedgradeFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    courseware_studentmodule = factory.RelatedFactoryList(
        EdxCoursewareStudentmoduleFactory,
        "student",
        size=3,
        student_id=factory.SelfAttribute("..id"),
    )
    courseware_xmodulestudentinfofield = factory.RelatedFactoryList(
        EdxCoursewareXmodulestudentinfofieldFactory,
        "student",
        size=3,
        student_id=factory.SelfAttribute("..id"),
    )
    courseware_xmodulestudentprefsfield = factory.RelatedFactoryList(
        EdxCoursewareXmodulestudentprefsfieldFactory,
        "student",
        size=3,
        student_id=factory.SelfAttribute("..id"),
    )
    django_comment_client_role_users = factory.RelatedFactoryList(
        EdxDjangoCommentClientRoleUsersFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    instructor_task_instructortask = factory.RelatedFactoryList(
        EdxInstructorTaskInstructortaskFactory,
        "requester",
        size=3,
        requester_id=factory.SelfAttribute("..id"),
    )
    notify_settings = factory.RelatedFactoryList(
        EdxNotifySettingFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    payment_useracceptance = factory.RelatedFactoryList(
        EdxPaymentUseracceptanceFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    proctoru_proctoruexam = factory.RelatedFactoryList(
        EdxProctoruProctoruexamFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    proctoru_proctoruuser = factory.RelatedFactory(
        EdxProctoruProctoruuserFactory, student_id=factory.SelfAttribute("..id")
    )
    student_anonymoususerid = factory.RelatedFactoryList(
        EdxStudentAnonymoususeridFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    student_courseaccessrole = factory.RelatedFactoryList(
        EdxStudentCourseaccessroleFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    student_courseenrollment = factory.RelatedFactoryList(
        EdxStudentCourseenrollmentFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    student_historicalcourseenrollment_history_user = factory.RelatedFactoryList(
        EdxStudentHistoricalcourseenrollmentFactory,
        "history_user",
        size=3,
        history_user_id=factory.SelfAttribute("..id"),
        user_id=factory.SelfAttribute("..id"),
    )
    student_historicalcourseenrollment_user = factory.RelatedFactoryList(
        EdxStudentHistoricalcourseenrollmentFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
        history_user_id=factory.SelfAttribute("..id"),
    )
    student_loginfailure = factory.RelatedFactoryList(
        EdxStudentLoginfailureFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    student_manualenrollmentaudit = factory.RelatedFactoryList(
        EdxStudentManualenrollmentauditFactory,
        "enrolled_by",
        size=3,
        enrolled_by_id=factory.SelfAttribute("..id"),
    )
    student_pendingemailchange = factory.RelatedFactory(
        EdxStudentPendingemailchangeFactory, user_id=factory.SelfAttribute("..id")
    )
    student_userstanding_user = factory.RelatedFactory(
        EdxStudentUserstandingFactory,
        changed_by_id=factory.SelfAttribute("..id"),
        user_id=factory.SelfAttribute("..id"),
    )
    user_api_userpreference = factory.RelatedFactoryList(
        EdxUserApiUserpreferenceFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
    )
    verify_student_softwaresecurephotoverification_reviewing_user = (
        factory.RelatedFactoryList(
            EdxVerifyStudentSoftwaresecurephotoverificationFactory,
            "reviewing_user",
            size=3,
            reviewing_user_id=factory.SelfAttribute("..id"),
            user_id=factory.SelfAttribute("..id"),
        )
    )
    verify_student_softwaresecurephotoverification_user = factory.RelatedFactoryList(
        EdxVerifyStudentSoftwaresecurephotoverificationFactory,
        "user",
        size=3,
        user_id=factory.SelfAttribute("..id"),
        reviewing_user_id=factory.SelfAttribute("..id"),
    )



"""Module for MongoDB CRUD functions."""

from logging import getLogger

from mongoengine.queryset.visitor import Q

from mork.conf import settings
from mork.edx.mongo.models import Comment, CommentThread

logger = getLogger(__name__)


def anonymize_comments(username: str) -> int:
    """Anonymize user comments and threads.

    Parameters:
    username (str): The username of the user to delete comments from.

    Returns the number of comments anonymized.
    """
    comments = Comment.objects(Q(author_username=username)).all()

    comment_count = comments.update(
        author_username="[deleted]",
        body="[deleted]",
        author_id=settings.EDX_FORUM_PLACEHOLDER_USER_ID,
        anonymous=True,
    )

    comment_threads = CommentThread.objects(Q(author_username=username)).all()

    comment_count += comment_threads.update(
        author_username="[deleted]",
        title="[deleted]",
        body="[deleted]",
        author_id=settings.EDX_FORUM_PLACEHOLDER_USER_ID,
        anonymous=True,
    )

    logger.info(f"Anonymized {comment_count} comment(s)")

    return comment_count



"""Mork Celery brevo tasks."""

from logging import getLogger
from uuid import UUID

import httpx

from mork.celery.celery_app import app
from mork.celery.utils import (
    get_service_status,
    get_user_from_mork,
    update_status_in_mork,
)
from mork.conf import settings
from mork.exceptions import (
    UserDeleteError,
    UserNotFound,
    UserStatusError,
)
from mork.models.users import DeletionStatus, ServiceName

logger = getLogger(__name__)


@app.task(
    bind=True,
    retry_kwargs={"max_retries": settings.DELETE_MAX_RETRIES},
)
def delete_brevo_platform_user(self, user_id: UUID):
    """Task to delete user from the Brevo platform."""
    if not settings.BREVO_API_URL:
        logger.info("Brevo API URL not set, skipping deletion.")
        return

    user = get_user_from_mork(user_id)
    if not user:
        msg = f"User {user_id} could not be retrieved from Mork"
        logger.error(msg)
        raise UserNotFound(msg)

    status = get_service_status(user, ServiceName.BREVO)

    if status == DeletionStatus.DELETED:
        logger.warning(f"User {user_id} has already been deleted.")
        return

    if status != DeletionStatus.TO_DELETE:
        msg = f"User {user_id} is not to be deleted. Status: {status}"
        logger.error(msg)
        raise UserStatusError(msg)

    try:
        delete_brevo_user(email=user.email)
    except UserDeleteError as exc:
        raise self.retry(exc=exc) from exc

    if not update_status_in_mork(
        user_id=user_id, service=ServiceName.BREVO, status=DeletionStatus.DELETED
    ):
        msg = f"Failed to update deletion status to deleted for user {user_id}"
        logger.error(msg)
        raise UserStatusError(msg)

    logger.info(f"Completed deletion process for user {user_id}")


def delete_brevo_user(email: str):
    """Delete user contact on Brevo."""
    logger.debug("Delete user contact on Brevo")

    try:
        response = httpx.delete(
            f"{settings.BREVO_API_URL}/contacts/{email}",
            headers={"api-key": f"{settings.BREVO_API_KEY}"},
        )
        response.raise_for_status()
    except httpx.RequestError as exc:
        msg = "Network error while deleting user contact on Brevo"
        logger.error(msg)
        raise UserDeleteError(msg) from exc
    except httpx.HTTPStatusError as exc:
        if exc.response.status_code == httpx.codes.NOT_FOUND:
            logger.info("User has no contact on Brevo")
            return

        msg = "Failed to delete user contact on Brevo"
        logger.error(msg)
        raise UserDeleteError(msg) from exc



"""Mork tasks models."""

from datetime import datetime
from uuid import uuid4

from sqlalchemy import DateTime, String
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import Mapped, mapped_column

from mork.models import Base


class EmailStatus(Base):
    """Model for storing email statuses."""

    __tablename__ = "email_status"

    filtered_attrs = ["email"]

    id: Mapped[int] = mapped_column(UUID(as_uuid=True), primary_key=True, default=uuid4)
    email: Mapped[str] = mapped_column(String(254), unique=True)
    sent_date: Mapped[datetime] = mapped_column(DateTime)



"""Mork edx notify models."""

from sqlalchemy import ForeignKeyConstraint
from sqlalchemy.dialects.mysql import INTEGER
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class NotifySetting(Base):
    """Model for the `notify_settings` table."""

    __tablename__ = "notify_settings"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    interval: Mapped[int] = mapped_column(INTEGER(6), nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="notify_settings"
    )



"""Factory classes for courseware models."""

import factory

from mork.edx.mysql.models.courseware import (
    CoursewareOfflinecomputedgrade,
    CoursewareStudentmodule,
    CoursewareStudentmodulehistory,
    CoursewareXmodulestudentinfofield,
    CoursewareXmodulestudentprefsfield,
)

from .base import BaseSQLAlchemyModelFactory, faker


class EdxCoursewareOfflinecomputedgradeFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `courseware_offlinecomputedgrade` table."""

    class Meta:
        """Factory configuration."""

        model = CoursewareOfflinecomputedgrade

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    created = factory.Faker("date_time")
    updated = factory.Faker("date_time")
    gradeset = factory.Faker("json")


class EdxCoursewareStudentmodulehistoryFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `courseware_studentmodulehistory` table."""

    class Meta:
        """Factory configuration."""

        model = CoursewareStudentmodulehistory

    id = factory.Sequence(lambda n: n + 1)
    version = factory.Faker("pystr")
    created = factory.Faker("date_time")
    state = factory.Faker("json")
    grade = factory.Faker("pyfloat")
    max_grade = factory.Faker("pyfloat")


class EdxCoursewareStudentmoduleFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `courseware_studentmodule` table."""

    class Meta:
        """Factory configuration."""

        model = CoursewareStudentmodule

    id = factory.Sequence(lambda n: n + 1)
    module_type = factory.Faker(
        "random_element", elements=["course", "sequential", "problem", "chapter"]
    )
    module_id = factory.Faker("pystr")
    student_id = factory.Sequence(lambda n: n + 1)
    state = factory.Faker("json")
    grade = factory.Faker("pyfloat")
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")
    max_grade = factory.Faker("pyfloat")
    done = factory.Faker("pystr", max_chars=8)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")

    courseware_studentmodulehistory = factory.RelatedFactoryList(
        EdxCoursewareStudentmodulehistoryFactory,
        "student_module",
        size=3,
        student_module_id=factory.SelfAttribute("..id"),
    )


class EdxCoursewareXmodulestudentinfofieldFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `courseware_xmodulestudentinfofield` table."""

    class Meta:
        """Factory configuration."""

        model = CoursewareXmodulestudentinfofield

    id = factory.Sequence(lambda n: n + 1)
    field_name = factory.Faker("pystr", max_chars=64)
    value = factory.Faker("random_element", elements=["true", "false"])
    student_id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")


class EdxCoursewareXmodulestudentprefsfieldFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `courseware_xmodulestudentprefsfield` table."""

    class Meta:
        """Factory configuration."""

        model = CoursewareXmodulestudentprefsfield

    id = factory.Sequence(lambda n: n + 1)
    field_name = factory.Faker("word")
    module_type = factory.Faker("word")
    value = factory.Faker("pystr")
    student_id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")



"""Mork edx user models."""

from sqlalchemy import ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class UserApiUserpreference(Base):
    """Model for the `user_api_userpreference` table."""

    __tablename__ = "user_api_userpreference"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
            name="user_id_refs_id_2839c1f4f3473b9e",
        ),
        Index("user_api_userpreference_45544485", "key"),
        Index("user_api_userpreference_fbfc09f1", "user_id"),
        Index(
            "user_api_userpreference_user_id_4e4942d73f760072_uniq",
            "user_id",
            "key",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11))
    key: Mapped[str] = mapped_column(String(255))
    value: Mapped[str] = mapped_column(TEXT)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="user_api_userpreference"
    )



"""Mork edx certificates models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class CertificatesCertificatehtmlviewconfiguration(Base):
    """Model for the `certificates_certificatehtmlviewconfiguration` table."""

    __tablename__ = "certificates_certificatehtmlviewconfiguration"
    __table_args__ = (
        ForeignKeyConstraint(
            ["changed_by_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    change_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    changed_by_id: Mapped[int] = mapped_column(INTEGER(11), index=True)
    enabled: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    configuration: Mapped[str] = mapped_column(TEXT, nullable=False)

    changed_by: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="certificates_certificatehtmlviewconfiguration"
    )


class CertificatesGeneratedcertificate(Base):
    """Model for the `certificates_generatedcertificate` table."""

    __tablename__ = "certificates_generatedcertificate"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
            name="user_id_refs_id_6c4fb3478e23bfe2",
        ),
        Index(
            "certificates_generatedcertific_verify_uuid_1b5a14bb83c471ff_uniq",
            "verify_uuid",
        ),
        Index(
            "certificates_generatedcertifica_course_id_1389f6b2d72f5e78_uniq",
            "course_id",
            "user_id",
            unique=True,
        ),
        Index("certificates_generatedcertificate_fbfc09f1", "user_id"),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    download_url: Mapped[str] = mapped_column(String(128), nullable=False)
    grade: Mapped[str] = mapped_column(String(5), nullable=False)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False)
    key: Mapped[str] = mapped_column(String(32), nullable=False)
    distinction: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    status: Mapped[str] = mapped_column(String(32), nullable=False)
    verify_uuid: Mapped[str] = mapped_column(String(32), nullable=False, index=True)
    download_uuid: Mapped[str] = mapped_column(String(32), nullable=False)
    name: Mapped[str] = mapped_column(String(255), nullable=False)
    created_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    modified_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    error_reason: Mapped[str] = mapped_column(String(512), nullable=False)
    mode: Mapped[str] = mapped_column(String(32), nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="certificates_generatedcertificate"
    )



"""Mork Celery edx tasks."""

from logging import getLogger
from uuid import UUID

from mongoengine.errors import OperationError
from sqlalchemy.exc import DBAPIError, SQLAlchemyError

from mork.celery.celery_app import app
from mork.celery.utils import (
    get_service_status,
    get_user_from_mork,
    update_status_in_mork,
)
from mork.conf import settings
from mork.edx.mongo import crud as mongo
from mork.edx.mongo.database import OpenEdxMongoDB
from mork.edx.mysql import crud as mysql
from mork.edx.mysql.database import OpenEdxMySQLDB
from mork.exceptions import (
    UserDeleteError,
    UserNotFound,
    UserProtected,
    UserStatusError,
)
from mork.models.users import DeletionStatus, ServiceName

logger = getLogger(__name__)


@app.task(
    bind=True,
    retry_kwargs={"max_retries": settings.DELETE_MAX_RETRIES},
)
def delete_edx_platform_user(self, user_id: UUID):
    """Task to delete user from the edX platform."""
    user = get_user_from_mork(user_id)
    if not user:
        msg = f"User {user_id} could not be retrieved from Mork"
        logger.error(msg)
        raise UserNotFound(msg)

    status = get_service_status(user, ServiceName.EDX)

    if status == DeletionStatus.PROTECTED:
        logger.warning(f"User {user_id} is protected.")
        return

    if status != DeletionStatus.TO_DELETE:
        msg = f"User {user_id} cannot be deleted. Status: {status}"
        logger.error(msg)
        raise UserStatusError(msg)

    new_status = DeletionStatus.DELETED
    try:
        delete_edx_mysql_user(email=user.email)
    except UserProtected:
        logger.info(f"User {user_id} is protected")
        new_status = DeletionStatus.PROTECTED
    except UserDeleteError as exc:
        logger.exception(exc)
        raise self.retry(exc=exc) from exc

    try:
        delete_edx_mongo_user(username=user.username)
    except UserDeleteError as exc:
        logger.exception(exc)
        raise self.retry(exc=exc) from exc

    if not update_status_in_mork(
        user_id=user_id, service=ServiceName.EDX, status=new_status
    ):
        msg = (
            f"Failed to update deletion status to '{new_status.value}' "
            f"for user {user_id}"
        )
        logger.error(msg)
        raise UserStatusError(msg)

    logger.info(f"Completed deletion process for user {user_id}")


def delete_edx_mysql_user(email: str):
    """Delete user's data from edX MySQL database."""
    logger.debug("Deleting user's data from edX MySQL")

    db = OpenEdxMySQLDB()
    try:
        mysql.delete_user(db.session, email=email)
        db.session.commit()
    except UserNotFound as exc:
        db.session.rollback()
        logger.info(f"Skipping MySQL deletion : {exc}")
    except UserProtected:
        db.session.rollback()
        db.session.close()
        raise
    except (SQLAlchemyError, DBAPIError) as exc:
        db.session.rollback()
        msg = "Failed to delete user from edX MySQL"
        logger.error(msg)
        raise UserDeleteError(msg) from exc
    finally:
        db.session.close()


def delete_edx_mongo_user(username: str):
    """Delete user's data from edX MongoDB database."""
    logger.debug("Deleting user's data from edX MongoDB")

    db = OpenEdxMongoDB()
    try:
        mongo.anonymize_comments(username)
    except OperationError as exc:
        msg = f"Failed to delete comments: {exc}"
        logger.error(msg)
        raise UserDeleteError(msg) from exc
    finally:
        db.disconnect()



"""Mork edx django models."""

from sqlalchemy import ForeignKeyConstraint, Index
from sqlalchemy.dialects.mysql import INTEGER
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class DjangoCommentClientRoleUsers(Base):
    """Model for the `django_comment_client_role_users` table."""

    __tablename__ = "django_comment_client_role_users"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "django_comment_client_role_users_role_id_78e483f531943614_uniq",
            "role_id",
            "user_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    role_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="django_comment_client_role_users"
    )



"""Tests for Mork Celery deletion tasks."""

import logging
from unittest.mock import Mock, call, patch

import pytest
from celery import group
from faker import Faker
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from mork.celery.tasks.deletion import (
    delete_inactive_users,
    delete_user,
    mark_user_for_deletion,
    remove_email_status,
)
from mork.conf import settings
from mork.edx.mysql.factories.auth import EdxAuthUserFactory
from mork.exceptions import UserDeleteError
from mork.factories.tasks import EmailStatusFactory
from mork.factories.users import UserFactory, UserServiceStatusFactory
from mork.models.tasks import EmailStatus
from mork.models.users import (
    DeletionReason,
    DeletionStatus,
    User,
)


def test_delete_inactive_users(edx_mysql_db, monkeypatch):
    """Test the `delete_inactive_users` function."""
    # 2 users that did not log in for more than the deletion period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe2@example.com",
    )
    # 2 users that logged in recently
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(start_date=-settings.DELETION_PERIOD),
        email="janedah1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(start_date=-settings.DELETION_PERIOD),
        email="janedah2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.group", mock_group)
    mock_delete_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.delete_user", mock_delete_user)

    delete_inactive_users(dry_run=False)

    mock_group.assert_called_once_with(
        [
            mock_delete_user.s(email="johndoe1@example.com", dry_run=False),
            mock_delete_user.s(email="johndoe2@example.com", dry_run=False),
        ]
    )


def test_delete_inactive_users_with_limit(edx_mysql_db, monkeypatch):
    """Test the `delete_inactive_users` function with limit."""
    # 2 users that did not log in for more than the deletion period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.group", mock_group)
    mock_delete_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.delete_user", mock_delete_user)

    delete_inactive_users(limit=1, dry_run=False)

    mock_group.assert_called_once_with(
        [
            mock_delete_user.s(email="johndoe1@example.com", dry_run=False),
        ]
    )


def test_delete_inactive_users_with_batch_size(edx_mysql_db, monkeypatch):
    """Test the `warn_inactive_users` function with batch size."""
    # 2 users that did not log in for more than the deletion period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.group", mock_group)
    mock_delete_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.delete_user", mock_delete_user)

    # Set batch size to 1
    monkeypatch.setattr(
        "mork.celery.tasks.deletion.settings.EDX_MYSQL_QUERY_BATCH_SIZE", 1
    )

    delete_inactive_users(dry_run=False)

    mock_group.assert_has_calls(
        [
            call(
                [
                    mock_delete_user.s(email="johndoe1@example.com", dry_run=False),
                ]
            ),
            call().delay(),
            call(
                [
                    mock_delete_user.s(email="johndoe2@example.com", dry_run=False),
                ]
            ),
            call().delay(),
        ]
    )


def test_delete_inactive_users_with_dry_run(edx_mysql_db, monkeypatch):
    """Test the `delete_inactive_users` function with dry run activated (by default)."""
    # 2 users that did not log in for more than the deletion period
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe1@example.com",
    )
    EdxAuthUserFactory.create(
        last_login=Faker().date_time_between(end_date=-settings.DELETION_PERIOD),
        email="johndoe2@example.com",
    )

    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    mock_group = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.group", mock_group)
    mock_delete_user = Mock()
    monkeypatch.setattr("mork.celery.tasks.deletion.delete_user", mock_delete_user)

    delete_inactive_users()

    mock_group.assert_called_once_with(
        [
            mock_delete_user.s(email="johndoe1@example.com", dry_run=True),
            mock_delete_user.s(email="johndoe2@example.com", dry_run=True),
        ]
    )


def test_delete_user(monkeypatch):
    """Test the `delete_user` function."""

    with (
        patch("mork.celery.tasks.deletion.remove_email_status.si") as mock_remove_email,
        patch("mork.celery.tasks.deletion.mark_user_for_deletion.si") as mock_mark,
        patch("mork.celery.tasks.deletion.delete_edx_platform_user.s") as mock_edx,
        patch("mork.celery.tasks.deletion.delete_brevo_platform_user.s") as mock_brevo,
        patch("mork.celery.tasks.deletion.chain") as mock_chain,
    ):
        email = "johndoe@example.com"
        reason = DeletionReason.USER_REQUESTED
        delete_user(email, reason, dry_run=False)

        mock_chain.assert_called_once_with(
            mock_remove_email(email=email),
            mock_mark(email=email, reason=reason),
            group(mock_edx(email=email), mock_brevo(email=email)),
        )


def test_delete_user_with_dry_run(monkeypatch):
    """Test the `delete_user` function with dry run activated (by default)."""

    with patch("mork.celery.tasks.deletion.chain") as mock_chain:
        email = "johndoe@example.com"
        delete_user(email)

        mock_chain.assert_not_called()


def test_mark_user_for_deletion(edx_mysql_db, db_session, caplog, monkeypatch):
    """Test the `mark_user_for_deletion` function."""

    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    class MockMorkDB:
        session = db_session

    monkeypatch.setattr("mork.celery.tasks.deletion.MorkDB", MockMorkDB)

    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    # As we are closing the session inside the function tested, the factory need to
    # commit the AuthUser created
    EdxAuthUserFactory._meta.sqlalchemy_session_persistence = "commit"
    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )
    auth_user = EdxAuthUserFactory.create()

    # Mark this user for deletion
    mark_user_for_deletion(auth_user.email, DeletionReason.GDPR)

    # Verify user is created
    inserted_user = db_session.scalars(select(User)).first()
    assert inserted_user.username == auth_user.username
    assert inserted_user.edx_user_id == auth_user.id
    assert inserted_user.email == auth_user.email
    assert inserted_user.reason == DeletionReason.GDPR
    # Verify statuses were created for all users and services
    assert all(
        status.status == DeletionStatus.TO_DELETE
        for status in inserted_user.service_statuses
    )

    # Try to mark the user for deletion again
    with caplog.at_level(logging.INFO):
        mark_user_for_deletion(auth_user.email, DeletionReason.GDPR)

    assert (
        "mork.celery.tasks.deletion",
        logging.INFO,
        "User is already marked for deletion",
    ) in caplog.record_tuples

    # Reset factory persistence
    EdxAuthUserFactory._meta.sqlalchemy_session_persistence = None


def test_mark_user_for_deletion_nonexistent_user(
    edx_mysql_db, db_session, caplog, monkeypatch
):
    """Test the `mark_user_for_deletion` function with a nonexistent user."""

    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )
    email = "johndoe@example.com"
    # Try to mark this user for deletion
    with pytest.raises(UserDeleteError, match="User not found in edx database"):
        mark_user_for_deletion(email, DeletionReason.GDPR)


def test_mark_user_for_deletion_read_failure(edx_mysql_db, db_session, monkeypatch):
    """Test the `mark_user_for_deletion` function with a read failure from MySQL."""

    def mock_get_user(*args, **kwargs):
        raise SQLAlchemyError("An error occurred")

    monkeypatch.setattr("mork.celery.tasks.deletion.crud.get_user", mock_get_user)

    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )
    auth_user = EdxAuthUserFactory.create()

    with pytest.raises(UserDeleteError, match="Failed to read user from edX MySQL"):
        mark_user_for_deletion(auth_user.email, reason=DeletionReason.GDPR)


def test_mark_user_for_deletion_write_failure(edx_mysql_db, db_session, monkeypatch):
    """Test the `mark_user_for_deletion` function with a write failure."""

    def mock_session_commit():
        raise SQLAlchemyError("An error occurred")

    db_session.commit = mock_session_commit

    class MockMorkDB:
        session = db_session

    monkeypatch.setattr("mork.celery.tasks.deletion.MorkDB", MockMorkDB)

    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    monkeypatch.setattr(
        "mork.celery.tasks.deletion.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )
    auth_user = EdxAuthUserFactory.create()

    with pytest.raises(UserDeleteError, match="Failed to mark user to be deleted"):
        mark_user_for_deletion(auth_user.email, reason=DeletionReason.GDPR)


def test_remove_email_status(db_session, monkeypatch):
    """Test the `remove_email_status` function."""

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.deletion.MorkDB", MockMorkDB)

    email = "johndoe1@example.com"
    EmailStatusFactory.create(email=email)

    # Check that an entry has been created for this email
    query = select(EmailStatus.email).where(EmailStatus.email == email)
    assert db_session.execute(query).scalars().first()

    # Delete entry
    remove_email_status(email)

    # Check that the entry has been deleted for this email
    query = select(EmailStatus.email).where(EmailStatus.email == email)
    assert not db_session.execute(query).scalars().first()


def test_remove_email_status_no_entry(caplog, db_session, monkeypatch):
    """Test the `remove_email_status` function when entry does not exist."""

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.deletion.MorkDB", MockMorkDB)

    email = "johndoe1@example.com"

    # Delete non existent entry
    with caplog.at_level(logging.WARNING):
        remove_email_status(email)

    assert (
        "mork.celery.tasks.deletion",
        logging.WARNING,
        "Email status not found",
    ) in caplog.record_tuples


def test_remove_email_status_with_failure(caplog, db_session, monkeypatch):
    """Test the `remove_email_status` with a commit failure."""

    def mock_session_commit():
        raise SQLAlchemyError("An error occurred")

    db_session.commit = mock_session_commit

    class MockMorkDB:
        session = db_session

    EmailStatusFactory._meta.sqlalchemy_session = db_session
    monkeypatch.setattr("mork.celery.tasks.deletion.MorkDB", MockMorkDB)

    email = "johndoe1@example.com"
    EmailStatusFactory.create(email=email)

    # Try to delete entry
    with caplog.at_level(logging.ERROR):
        remove_email_status(email)

    assert (
        "mork.celery.tasks.deletion",
        logging.ERROR,
        "Failed to delete email status",
    ) in caplog.record_tuples



"""Factory classes for django models."""

import factory

from mork.edx.mysql.models.django import DjangoCommentClientRoleUsers

from .base import BaseSQLAlchemyModelFactory


class EdxDjangoCommentClientRoleUsersFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `django_comment_client_role_users` table."""

    class Meta:
        """Factory configuration."""

        model = DjangoCommentClientRoleUsers

    id = factory.Sequence(lambda n: n + 1)
    role_id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)



"""Tests for the health check endpoints."""

import pytest


@pytest.mark.anyio
async def test_api_health_lbheartbeat(http_client):
    """Test the load balancer heartbeat healthcheck."""
    response = await http_client.get("/__lbheartbeat__")
    assert response.status_code == 200
    assert response.json() is None


@pytest.mark.anyio
async def test_api_health_heartbeat(db_session, http_client, monkeypatch):
    """Test the heartbeat healthcheck."""
    monkeypatch.setattr("mork.db.get_session", db_session)

    response = await http_client.get("/__heartbeat__")
    assert response.status_code == 200
    assert response.json() == {"database": "ok"}

    monkeypatch.setattr("mork.api.health.is_db_alive", lambda x: False)

    response = await http_client.get("/__heartbeat__")
    assert response.status_code == 500
    assert response.json() == {"database": "error"}



"""Factory classes for verify models."""

import factory

from mork.edx.mysql.models.verify import (
    VerifyStudentHistoricalverificationdeadline,
    VerifyStudentSoftwaresecurephotoverification,
)

from .base import BaseSQLAlchemyModelFactory, faker


class EdxVerifyStudentHistoricalverificationdeadlineFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `verify_student_historicalverificationdeadline` table."""

    class Meta:
        """Factory configuration."""

        model = VerifyStudentHistoricalverificationdeadline

    id = factory.Sequence(lambda n: n + 1)
    created = factory.Faker("date_time")
    modified = factory.Faker("date_time")
    course_key = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    deadline = factory.Faker("date_time")
    history_id = factory.Sequence(lambda n: n + 1)
    history_date = factory.Faker("date_time")
    history_user_id = factory.Sequence(lambda n: n + 1)
    history_type = factory.Faker("random_element", elements=["+", "-", "~"])
    deadline_is_explicit = factory.Faker("random_int", min=0, max=1)


class EdxVerifyStudentSoftwaresecurephotoverificationFactory(
    BaseSQLAlchemyModelFactory
):
    """Factory for the `verify_student_softwaresecurephotoverification` table."""

    class Meta:
        """Factory configuration."""

        model = VerifyStudentSoftwaresecurephotoverification

    id = factory.Sequence(lambda n: n + 1)
    status = factory.Faker("pystr")
    status_changed = factory.Faker("date_time")
    user_id = factory.Sequence(lambda n: n + 1)
    name = factory.Faker("pystr")
    face_image_url = factory.Faker("url")
    photo_id_image_url = factory.Faker("url")
    receipt_id = factory.Faker("pystr")
    created_at = factory.Faker("date_time")
    updated_at = factory.Faker("date_time")
    submitted_at = factory.Faker("date_time")
    reviewing_user_id = factory.Sequence(lambda n: n + 1)
    reviewing_service = factory.Faker("pystr")
    error_msg = factory.Faker("text")
    error_code = factory.Faker("pystr")
    photo_id_key = factory.Faker("text")
    display = factory.Faker("random_int", min=0, max=1)
    copy_id_photo_from_id = factory.Sequence(lambda n: n + 1)



"""Mork edx bulk models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class BulkEmailCourseemail(Base):
    """Model for the `bulk_email_courseemail` table."""

    __tablename__ = "bulk_email_courseemail"
    __table_args__ = (
        ForeignKeyConstraint(
            ["sender_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    sender_id: Mapped[int] = mapped_column(INTEGER(11))
    slug: Mapped[str] = mapped_column(String(128), nullable=False, index=True)
    subject: Mapped[str] = mapped_column(String(128), nullable=False)
    html_message: Mapped[str] = mapped_column(TEXT)
    created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    modified: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    to_option: Mapped[str] = mapped_column(String(64), nullable=False)
    text_message: Mapped[str] = mapped_column(TEXT)
    template_name: Mapped[str] = mapped_column(String(255))
    from_addr: Mapped[str] = mapped_column(String(255))

    sender: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="bulk_email_courseemail"
    )


class BulkEmailOptout(Base):
    """Model for the `bulk_email_optout` table."""

    __tablename__ = "bulk_email_optout"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "bulk_email_optout_course_id_368f7519b2997e1a_uniq",
            "course_id",
            "user_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    course_id: Mapped[str] = mapped_column(String(255), nullable=False, index=True)
    user_id: Mapped[str] = mapped_column(INTEGER(11), index=True)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="bulk_email_optout"
    )



"""Factory classes for certificates models."""

import factory

from mork.edx.mysql.models.certificates import (
    CertificatesCertificatehtmlviewconfiguration,
    CertificatesGeneratedcertificate,
)

from .base import BaseSQLAlchemyModelFactory, faker


class EdxCertificatesCertificatehtmlviewconfigurationFactory(
    BaseSQLAlchemyModelFactory
):
    """Factory for the `certificates_certificatehtmlviewconfiguration` table."""

    class Meta:
        """Factory configuration."""

        model = CertificatesCertificatehtmlviewconfiguration

    id = factory.Sequence(lambda n: n + 1)
    change_date = factory.Faker("date_time")
    changed_by_id = factory.Sequence(lambda n: n + 1)
    enabled = True
    configuration = factory.Faker("pystr")


class EdxCertificatesGeneratedCertificateFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `certificates_generatedcertificate` table."""

    class Meta:
        """Factory configuration."""

        model = CertificatesGeneratedcertificate

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    download_url = factory.Faker("uri")
    grade = factory.LazyAttribute(lambda _: str(faker.pyfloat())[:5])
    key = factory.Faker("hexify", text="^" * 32)
    distinction = factory.Faker("pyint")
    status = factory.Faker(
        "random_element", elements=["downloadable", "notpassing", "unavailable"]
    )
    verify_uuid = factory.Faker("hexify", text="^" * 32)
    download_uuid = factory.Faker("hexify", text="^" * 32)
    name = factory.Faker("name")
    created_date = factory.Faker("date_time")
    modified_date = factory.Faker("date_time")
    error_reason = factory.Faker("pystr")
    mode = factory.Faker("word")



"""Mork edx proctoru models."""

import datetime

from sqlalchemy import DateTime, ForeignKeyConstraint, Index, String
from sqlalchemy.dialects.mysql import INTEGER, TEXT
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Base


class ProctoruProctoruexam(Base):
    """Model for the `proctoru_proctoruexam` table."""

    __tablename__ = "proctoru_proctoruexam"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["auth_user.id"],
        ),
        Index(
            "proctoru_proctoruexam_user_id_2d5ce6ea053f2d80_uniq",
            "user_id",
            "block_id",
            unique=True,
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    user_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, index=True)
    start_date: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    actual_start_time: Mapped[datetime.datetime] = mapped_column(DateTime)
    is_completed: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    is_started: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    is_canceled: Mapped[int] = mapped_column(INTEGER(1), nullable=False)
    block_id: Mapped[str] = mapped_column(String(200), nullable=False)
    end_time: Mapped[datetime.datetime] = mapped_column(DateTime)
    reservation_id: Mapped[str] = mapped_column(String(50), nullable=False)
    reservation_no: Mapped[str] = mapped_column(String(200), nullable=False)
    url: Mapped[str] = mapped_column(TEXT, nullable=False)

    user: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="proctoru_proctoruexam"
    )


class ProctoruProctoruuser(Base):
    """Model for the `proctoru_proctoruuser` table."""

    __tablename__ = "proctoru_proctoruuser"
    __table_args__ = (
        ForeignKeyConstraint(
            ["student_id"],
            ["auth_user.id"],
        ),
    )

    id: Mapped[int] = mapped_column(INTEGER(11), primary_key=True)
    student_id: Mapped[int] = mapped_column(INTEGER(11), nullable=False, unique=True)
    phone_number: Mapped[str] = mapped_column(String(100), nullable=False)
    time_zone: Mapped[str] = mapped_column(String(100), nullable=False)
    address: Mapped[str] = mapped_column(TEXT, nullable=False)
    city: Mapped[str] = mapped_column(String(50), nullable=False)
    country: Mapped[str] = mapped_column(String(50), nullable=False)
    date_created: Mapped[datetime.datetime] = mapped_column(DateTime, nullable=False)
    time_zone_display_name: Mapped[str] = mapped_column(String(100), nullable=False)
    state: Mapped[str] = mapped_column(String(50), nullable=False)

    student: Mapped["AuthUser"] = relationship(  # noqa: F821
        "AuthUser", back_populates="proctoru_proctoruuser"
    )



"""Edx database test fixtures."""

import mongomock
import pytest
from alembic import command
from alembic.config import Config
from mongoengine import connect, disconnect
from sqlalchemy import create_engine
from sqlalchemy.orm import Session as SASession

from mork.api.v1 import app as v1
from mork.conf import settings
from mork.db import get_session
from mork.edx.mongo.database import OpenEdxMongoDB
from mork.edx.mysql.database import OpenEdxMySQLDB
from mork.edx.mysql.factories.base import Session, engine
from mork.edx.mysql.models.base import Base as EdxBase
from mork.models.tasks import Base


@pytest.fixture
def edx_mysql_db():
    """Test edx MySQL database fixture."""
    Session.configure(bind=engine)
    db = OpenEdxMySQLDB(engine, Session)
    EdxBase.metadata.create_all(engine)
    yield db
    db.session.rollback()
    EdxBase.metadata.drop_all(engine)


@pytest.fixture
def edx_mongo_db():
    """Test edx MongoDB database fixture."""
    connection = connect(
        host=settings.EDX_MONGO_DB_HOST,
        db=settings.EDX_MONGO_DB_NAME,
        mongo_client_class=mongomock.MongoClient,
    )
    db = OpenEdxMongoDB(connection)
    yield db
    disconnect()


@pytest.fixture(scope="session")
def db_engine():
    """Test database engine fixture."""
    engine = create_engine(settings.TEST_DB_URL, echo=False)
    # Create database and tables
    Base.metadata.create_all(engine)

    # Pretend to have all migrations applied
    alembic_cfg = Config(settings.ALEMBIC_CFG_PATH)
    alembic_cfg.set_main_option("sqlalchemy.url", settings.TEST_DB_URL)
    command.stamp(alembic_cfg, "head")

    yield engine

    Base.metadata.drop_all(engine)
    engine.dispose()


@pytest.fixture(scope="function")
def db_session(db_engine):
    """Test session fixture."""
    # Setup
    #
    # Connect to the database and create a non-ORM transaction. Our connection
    # is bound to the test session.
    connection = db_engine.connect()
    transaction = connection.begin()
    session = SASession(bind=connection)

    yield session

    # Teardown
    #
    # Rollback everything that happened with the Session above (including
    # explicit commits).
    session.close()
    transaction.rollback()
    connection.close()


@pytest.fixture(autouse=True)
def override_db_test_session(db_session):
    """Use test database along with a test session by default."""

    def get_session_override():
        return db_session

    v1.dependency_overrides[get_session] = get_session_override

    yield



"""Factory classes for course models."""

import factory

from mork.edx.mysql.models.course import (
    CourseActionStateCoursererunstate,
    CourseCreatorsCoursecreator,
    CourseGroupsCohortmembership,
    CourseGroupsCourseusergroupUsers,
)

from .base import BaseSQLAlchemyModelFactory, faker


class EdxCourseActionStateCoursererunstateFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `course_action_state_coursererunstate` table."""

    class Meta:
        """Factory configuration."""

        model = CourseActionStateCoursererunstate

    id = factory.Sequence(lambda n: n + 1)
    created_time = factory.Faker("date_time")
    updated_time = factory.Faker("date_time")
    course_key = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    action = factory.Faker("word")
    state = factory.Faker("word")
    should_display = factory.Faker("random_int", min=0, max=1)
    message = factory.Faker("text")
    source_course_key = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")
    display_name = factory.Faker("text")


class EdxCourseCreatorsCoursecreatorFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `course_creators_coursecreator` table."""

    class Meta:
        """Factory configuration."""

        model = CourseCreatorsCoursecreator

    id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    state_changed = factory.Faker("date_time")
    state = factory.Faker("word")
    note = factory.Faker("text")


class EdxCourseGroupsCourseusergroupUsersFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `course_groups_courseusergroup_users` table."""

    class Meta:
        """Factory configuration."""

        model = CourseGroupsCourseusergroupUsers

    id = factory.Sequence(lambda n: n + 1)
    courseusergroup_id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)


class EdxCourseGroupsCohortmembershipFactory(BaseSQLAlchemyModelFactory):
    """Factory for the `course_groups_cohortmembership` table."""

    class Meta:
        """Factory configuration."""

        model = CourseGroupsCohortmembership

    id = factory.Sequence(lambda n: n + 1)
    course_user_group_id = factory.Sequence(lambda n: n + 1)
    user_id = factory.Sequence(lambda n: n + 1)
    course_id = factory.Sequence(lambda n: f"course-v1:edX+{faker.pystr()}+{n}")



"""Tests for Mork Celery edx tasks."""

import logging
from unittest.mock import Mock
from uuid import uuid4

import pytest
from mongoengine.errors import OperationError
from mongoengine.queryset.visitor import Q
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from mork.celery.tasks.edx import (
    delete_edx_mongo_user,
    delete_edx_mysql_user,
    delete_edx_platform_user,
)
from mork.edx.mongo.factories import CommentFactory, CommentThreadFactory
from mork.edx.mongo.models import Comment, CommentThread
from mork.edx.mysql import crud as mysql
from mork.edx.mysql.factories.auth import EdxAuthUserFactory
from mork.exceptions import (
    UserDeleteError,
    UserNotFound,
    UserProtected,
    UserStatusError,
)
from mork.factories.users import UserFactory, UserServiceStatusFactory
from mork.models.users import DeletionStatus, ServiceName, User
from mork.schemas.users import UserRead


def test_delete_edx_platform_user(db_session, monkeypatch):
    """Test to delete user from edX platform."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    delete_edx_platform_user(user.id)

    mock_delete_edx_mysql_user.assert_called_once_with(email=user.email)
    mock_delete_edx_mongo_user.assert_called_once_with(username=user.username)
    mock_update_status_in_mork.assert_called_once_with(
        user_id=user.id, service=ServiceName.EDX, status=DeletionStatus.DELETED
    )


def test_delete_edx_platform_user_protected(db_session, monkeypatch):
    """Test to delete user and update its status to protected."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock(side_effect=UserProtected())
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    delete_edx_platform_user(user.id)

    mock_delete_edx_mysql_user.assert_called_once_with(email=user.email)
    mock_delete_edx_mongo_user.assert_called_once_with(username=user.username)
    mock_update_status_in_mork.assert_called_once_with(
        user_id=user.id, service=ServiceName.EDX, status=DeletionStatus.PROTECTED
    )


def test_delete_edx_platform_user_invalid_user(monkeypatch):
    """Test to delete user from edX platform with an invalid user."""

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: None)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    nonexistent_id = uuid4().hex
    with pytest.raises(
        UserNotFound, match=f"User {nonexistent_id} could not be retrieved from Mork"
    ):
        delete_edx_platform_user(nonexistent_id)

    mock_delete_edx_mysql_user.assert_not_called()
    mock_delete_edx_mongo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_edx_platform_user_protected_status(db_session, monkeypatch):
    """Test to delete user from edX platform with a status PROTECTED."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is protected on edx
    UserFactory.create(
        service_statuses={ServiceName.EDX: DeletionStatus.PROTECTED},
    )

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    delete_edx_platform_user(user.id)

    mock_delete_edx_mysql_user.assert_not_called()
    mock_delete_edx_mongo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_edx_platform_user_invalid_status(db_session, monkeypatch):
    """Test to delete user from edX platform with an invalid status."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is protected on edx
    UserFactory.create(
        service_statuses={ServiceName.EDX: DeletionStatus.DELETED},
    )

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=True)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    with pytest.raises(
        UserStatusError,
        match=f"User {user.id} cannot be deleted. Status: DeletionStatus.DELETED",
    ):
        delete_edx_platform_user(user.id)

    mock_delete_edx_mysql_user.assert_not_called()
    mock_delete_edx_mongo_user.assert_not_called()
    mock_update_status_in_mork.assert_not_called()


def test_delete_edx_platform_user_failed_delete(db_session, monkeypatch):
    """Test to delete user from edX platform with a failed delete."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is already deleted on edx
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock(side_effect=UserDeleteError("An error occurred"))
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )

    with pytest.raises(UserDeleteError, match="An error occurred"):
        delete_edx_platform_user(user.id)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock(side_effect=UserDeleteError("An error occurred"))
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )

    with pytest.raises(UserDeleteError, match="An error occurred"):
        delete_edx_platform_user(user.id)


def test_delete_edx_platform_user_failed_status_update(db_session, monkeypatch):
    """Test to delete user from edX platform with a failed status update."""
    UserServiceStatusFactory._meta.sqlalchemy_session = db_session
    UserFactory._meta.sqlalchemy_session = db_session

    # Create one user in the database that is already deleted on edx
    UserFactory.create()

    # Get user from db
    user = UserRead.model_validate(db_session.scalar(select(User)))

    monkeypatch.setattr("mork.celery.tasks.edx.get_user_from_mork", lambda x: user)

    mock_delete_edx_mysql_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mysql_user", mock_delete_edx_mysql_user
    )
    mock_delete_edx_mongo_user = Mock()
    monkeypatch.setattr(
        "mork.celery.tasks.edx.delete_edx_mongo_user", mock_delete_edx_mongo_user
    )
    mock_update_status_in_mork = Mock(return_value=False)
    monkeypatch.setattr(
        "mork.celery.tasks.edx.update_status_in_mork", mock_update_status_in_mork
    )

    with pytest.raises(
        UserStatusError,
        match=f"Failed to update deletion status to 'deleted' for user {user.id}",
    ):
        delete_edx_platform_user(user.id)


def test_delete_edx_mysql_user(edx_mysql_db, monkeypatch):
    """Test to delete user's data from MySQL."""
    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    EdxAuthUserFactory.create(email="johndoe1@example.com")
    EdxAuthUserFactory.create(email="johndoe2@example.com")

    monkeypatch.setattr(
        "mork.celery.tasks.edx.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    # Check both users exist on the MySQL database
    assert mysql.get_user(
        edx_mysql_db.session,
        email="johndoe1@example.com",
    )
    assert mysql.get_user(
        edx_mysql_db.session,
        email="johndoe2@example.com",
    )

    delete_edx_mysql_user(email="johndoe1@example.com")

    # Check only one remains
    assert not mysql.get_user(
        edx_mysql_db.session,
        email="johndoe1@example.com",
    )
    assert mysql.get_user(
        edx_mysql_db.session,
        email="johndoe2@example.com",
    )


def test_delete_edx_mysql_user_protected(edx_mysql_db, monkeypatch):
    """Test to delete data from MySQL for a protected user."""
    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    email = "johndoe1@example.com"
    EdxAuthUserFactory.create(email=email)

    def mock_delete_user(*args, **kwargs):
        raise UserProtected("An error occurred")

    monkeypatch.setattr("mork.celery.tasks.edx.mysql.delete_user", mock_delete_user)

    with pytest.raises(UserProtected, match="An error occurred"):
        delete_edx_mysql_user(email=email)

    # Check the user still exists on the edX MySQL database
    assert mysql.get_user(
        edx_mysql_db.session,
        email="johndoe1@example.com",
    )


def test_delete_edx_mysql_user_not_found(edx_mysql_db, monkeypatch, caplog):
    """Test that deleting a nonexistent user from MySQL should silently return."""

    monkeypatch.setattr(
        "mork.celery.tasks.edx.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    with caplog.at_level(logging.INFO):
        delete_edx_mysql_user(email="johndoe@example.com")

    assert (
        "mork.celery.tasks.edx",
        logging.INFO,
        "Skipping MySQL deletion : User does not exist",
    ) in caplog.record_tuples


def test_delete_edx_mysql_user_with_failure(edx_mysql_db, monkeypatch):
    """Test to delete user's data from MySQL with a commit failure."""
    EdxAuthUserFactory._meta.sqlalchemy_session = edx_mysql_db.session
    email = "johndoe1@example.com"
    EdxAuthUserFactory.create(email=email)

    def mock_session_commit():
        raise SQLAlchemyError("An error occurred")

    edx_mysql_db.session.commit = mock_session_commit
    monkeypatch.setattr(
        "mork.celery.tasks.edx.OpenEdxMySQLDB", lambda *args: edx_mysql_db
    )

    with pytest.raises(
        UserDeleteError,
        match="Failed to delete user from edX MySQL",
    ):
        delete_edx_mysql_user(email=email)


def test_delete_edx_mongo_user(edx_mongo_db, monkeypatch):
    """Test to delete user's data from MongoDB."""
    # Create one comment and one thread for two different users
    user_1 = "Johndoe1"
    CommentFactory.create(author_username=user_1)
    CommentThreadFactory.create(author_username=user_1)
    user_2 = "Johndoe2"
    CommentFactory.create(author_username=user_2)
    CommentThreadFactory.create(author_username=user_2)

    edx_mongo_db.disconnect = lambda: None
    monkeypatch.setattr(
        "mork.celery.tasks.edx.OpenEdxMongoDB", lambda *args: edx_mongo_db
    )

    # Check both users have comments in the MongoDB database
    assert Comment.objects(Q(author_username=user_1)).all().count() == 1
    assert CommentThread.objects(Q(author_username=user_1)).all().count() == 1
    assert Comment.objects(Q(author_username=user_2)).all().count() == 1
    assert CommentThread.objects(Q(author_username=user_2)).all().count() == 1

    delete_edx_mongo_user(username=user_1)

    # Check only one user remains
    assert Comment.objects(Q(author_username=user_1)).all().count() == 0
    assert CommentThread.objects(Q(author_username=user_1)).all().count() == 0
    assert Comment.objects(Q(author_username=user_2)).all().count() == 1
    assert CommentThread.objects(Q(author_username=user_2)).all().count() == 1


def test_delete_edx_mongo_user_with_failure(edx_mongo_db, monkeypatch):
    """Test to delete user's data from MongoDB with a operation failure."""
    username = "Johndoe1"
    CommentFactory.create(author_username=username)
    CommentThreadFactory.create(author_username=username)

    monkeypatch.setattr(
        "mork.celery.tasks.edx.OpenEdxMongoDB", lambda *args: edx_mongo_db
    )

    def mock_anonymize(*args):
        raise OperationError("An error occurred")

    monkeypatch.setattr(
        "mork.celery.tasks.edx.mongo.anonymize_comments", mock_anonymize
    )

    with pytest.raises(
        UserDeleteError,
        match="Failed to delete comments: An error occurred",
    ):
        delete_edx_mongo_user(username=username)



"""Module py.test fixtures."""

# ruff: noqa: F401

from pathlib import Path

from .fixtures.app import http_client
from .fixtures.asynchronous import anyio_backend
from .fixtures.auth import auth_headers
from .fixtures.db import (
    db_engine,
    db_session,
    edx_mongo_db,
    edx_mysql_db,
    override_db_test_session,
)

TEST_STATIC_PATH = Path(__file__).parent / "static"
